@@ -87,6 +87,7 @@ struct session_req {
 struct service_auth {
 	service_auth_cb cb;
 	void *user_data;
+	struct btd_device *device;
 };
 
 struct btd_adapter {
@@ -205,6 +206,7 @@ static int found_device_cmp(const struct remote_dev_info *d1,
 
 static void dev_info_free(struct remote_dev_info *dev)
 {
+	g_free(dev->name);
 	g_free(dev->alias);
 	g_free(dev);
 }
@@ -233,9 +235,9 @@ int pending_remote_name_cancel(struct btd_adapter *adapter)
 
 	if (hci_read_remote_name_cancel(dd, &dev->bdaddr,
 					HCI_REQ_TIMEOUT) < 0) {
-		error("Remote name cancel failed: %s(%d)", strerror(errno),
-									errno);
 		err = -errno;
+		error("Remote name cancel failed: %s(%d)",
+						strerror(errno), errno);
 	}
 
 	/* free discovered devices list */
@@ -261,7 +263,7 @@ static int set_limited_discoverable(int dd, const uint8_t *cls,
 	if (hci_write_current_iac_lap(dd, num, lap, HCI_REQ_TIMEOUT) < 0) {
 		int err = -errno;
 		error("Can't write current IAC LAP: %s(%d)",
-				strerror(err), err);
+						strerror(errno), errno);
 		return err;
 	}
 
@@ -280,7 +282,7 @@ static int set_limited_discoverable(int dd, const uint8_t *cls,
 	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
 		int err = -errno;
 		error("Can't write class of device: %s (%d)",
-							strerror(err), err);
+						strerror(errno), errno);
 		return err;
 	}
 
@@ -858,14 +860,14 @@ static int adapter_set_name(struct btd_adapter *adapter, const char *name)
 	if (dd < 0) {
 		err = -errno;
 		error("Can't open device hci%d: %s (%d)",
-					adapter->dev_id, strerror(err), err);
+				adapter->dev_id, strerror(errno), errno);
 		return err;
 	}
 
 	if (hci_write_local_name(dd, name, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't write name for hci%d: %s (%d)",
-					adapter->dev_id, strerror(err), err);
+				adapter->dev_id, strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -995,6 +997,13 @@ void adapter_remove_device(DBusConnection *conn, struct btd_adapter *adapter,
 			DBUS_TYPE_OBJECT_PATH, &dev_path,
 			DBUS_TYPE_INVALID);
 
+	agent = device_get_agent(device);
+	if (!agent)
+		agent = adapter->agent;
+
+	if (agent && device_is_authorizing(device))
+		agent_cancel(agent);
+
 	agent = device_get_agent(device);
 
 	if (agent) {
@@ -1054,7 +1063,7 @@ static int start_inquiry(struct btd_adapter *adapter)
 	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Unable to start inquiry: %s (%d)",
-			strerror(err), err);
+						strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -1104,7 +1113,7 @@ static int start_periodic_inquiry(struct btd_adapter *adapter)
 	if (hci_send_req(dd, &rq, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Unable to start periodic inquiry: %s (%d)",
-				strerror(err), err);
+						strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -1393,7 +1402,7 @@ static DBusMessage *request_session(DBusConnection *conn,
 		return dbus_message_new_method_return(msg);
 
 	ret = agent_confirm_mode_change(adapter->agent, mode2str(new_mode),
-					confirm_mode_cb, req);
+					confirm_mode_cb, req, NULL);
 	if (ret < 0) {
 		session_unref(req);
 		return failed_strerror(msg, -ret);
@@ -1469,12 +1478,12 @@ static DBusMessage *cancel_device_creation(DBusConnection *conn,
 		return invalid_args(msg);
 
 	device = adapter_find_device(adapter, address);
-	if (!device)
+	if (!device || !device_is_temporary(device))
 		return g_dbus_create_error(msg,
 				ERROR_INTERFACE ".NotInProgress",
 				"Device creation not in progress");
 
-	if (!device_is_temporary(device) || !device_is_bonding(device, sender))
+	if (!device_is_bonding(device, sender))
 		return not_authorized(msg);
 
 	adapter_remove_device(conn, adapter, device);
@@ -1770,14 +1779,14 @@ static int adapter_read_bdaddr(uint16_t dev_id, bdaddr_t *bdaddr)
 	if (dd < 0) {
 		err = -errno;
 		error("Can't open device hci%d: %s (%d)",
-					dev_id, strerror(err), err);
+					dev_id, strerror(errno), errno);
 		return err;
 	}
 
 	if (hci_read_bd_addr(dd, bdaddr, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't read address for hci%d: %s (%d)",
-					dev_id, strerror(err), err);
+					dev_id, strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -1842,7 +1851,7 @@ static int adapter_setup(struct btd_adapter *adapter, int dd)
 	if (hci_write_inquiry_mode(dd, inqmode, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't write inquiry mode for %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -2112,14 +2121,14 @@ int adapter_start(struct btd_adapter *adapter)
 	if (dd < 0) {
 		err = -errno;
 		error("Can't open adapter %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		return err;
 	}
 
 	if (hci_read_local_version(dd, &ver, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't read version info for %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -2132,7 +2141,7 @@ int adapter_start(struct btd_adapter *adapter)
 	if (hci_read_local_features(dd, features, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't read features for %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -2142,7 +2151,7 @@ int adapter_start(struct btd_adapter *adapter)
 	if (hci_read_class_of_dev(dd, dev->class, HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't read class of adapter on %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -2151,7 +2160,7 @@ int adapter_start(struct btd_adapter *adapter)
 						HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't read local name on %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		hci_close_dev(dd);
 		return err;
 	}
@@ -2168,7 +2177,7 @@ int adapter_start(struct btd_adapter *adapter)
 						HCI_REQ_TIMEOUT) < 0) {
 		err = -errno;
 		error("Can't read simple pairing mode on %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		/* Fall through since some chips have broken
 		 * read_simple_pairing_mode behavior */
 	}
@@ -2271,9 +2280,9 @@ int adapter_stop(struct btd_adapter *adapter)
 		adapter->oor_devices = NULL;
 	}
 
-	if (adapter->connections) {
-		g_slist_free(adapter->connections);
-		adapter->connections = NULL;
+	while (adapter->connections) {
+		struct btd_device *device = adapter->connections->data;
+		adapter_remove_connection(adapter, device, 0);
 	}
 
 	if (adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY)) {
@@ -2328,7 +2337,7 @@ int adapter_update(struct btd_adapter *adapter, uint8_t new_svc,
 	if (dd < 0) {
 		int err = -errno;
 		error("Can't open adapter %s: %s (%d)",
-					adapter->path, strerror(err), err);
+					adapter->path, strerror(errno), errno);
 		return err;
 	}
 
@@ -2534,8 +2543,100 @@ static int dev_rssi_cmp(struct remote_dev_info *d1, struct remote_dev_info *d2)
 	return rssi1 - rssi2;
 }
 
-int adapter_add_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr,
-				int8_t rssi, uint32_t class, const char *alias,
+static void append_dict_valist(DBusMessageIter *iter,
+					const char *first_key,
+					va_list var_args)
+{
+	DBusMessageIter dict;
+	const char *key;
+	int type;
+	void *val;
+
+	dbus_message_iter_open_container(iter, DBUS_TYPE_ARRAY,
+			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
+			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
+			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+
+	key = first_key;
+	while (key) {
+		type = va_arg(var_args, int);
+		val = va_arg(var_args, void *);
+		dict_append_entry(&dict, key, type, val);
+		key = va_arg(var_args, char *);
+	}
+
+	dbus_message_iter_close_container(iter, &dict);
+}
+
+static void emit_device_found(const char *path, const char *address,
+				const char *first_key, ...)
+{
+	DBusMessage *signal;
+	DBusMessageIter iter;
+	va_list var_args;
+
+	signal = dbus_message_new_signal(path, ADAPTER_INTERFACE,
+					"DeviceFound");
+	if (!signal) {
+		error("Unable to allocate new %s.DeviceFound signal",
+				ADAPTER_INTERFACE);
+		return;
+	}
+	dbus_message_iter_init_append(signal, &iter);
+	dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &address);
+
+	va_start(var_args, first_key);
+	append_dict_valist(&iter, first_key, var_args);
+	va_end(var_args);
+
+	g_dbus_send_message(connection, signal);
+}
+
+void adapter_emit_device_found(struct btd_adapter *adapter,
+				struct remote_dev_info *dev)
+{
+	struct btd_device *device;
+	char peer_addr[18], local_addr[18];
+	const char *icon, *paddr = peer_addr;
+	dbus_bool_t paired = FALSE;
+	dbus_int16_t rssi = dev->rssi;
+	char *alias;
+
+	ba2str(&dev->bdaddr, peer_addr);
+	ba2str(&adapter->bdaddr, local_addr);
+
+	device = adapter_find_device(adapter, paddr);
+	if (device)
+		paired = device_is_paired(device);
+
+	icon = class_to_icon(dev->class);
+
+	if (!dev->alias) {
+		if (!dev->name) {
+			alias = g_strdup(peer_addr);
+			g_strdelimit(alias, ":", '-');
+		} else
+			alias = g_strdup(dev->name);
+	} else
+		alias = g_strdup(dev->alias);
+
+	emit_device_found(adapter->path, paddr,
+			"Address", DBUS_TYPE_STRING, &paddr,
+			"Class", DBUS_TYPE_UINT32, &dev->class,
+			"Icon", DBUS_TYPE_STRING, &icon,
+			"RSSI", DBUS_TYPE_INT16, &rssi,
+			"Name", DBUS_TYPE_STRING, &dev->name,
+			"Alias", DBUS_TYPE_STRING, &alias,
+			"LegacyPairing", DBUS_TYPE_BOOLEAN, &dev->legacy,
+			"Paired", DBUS_TYPE_BOOLEAN, &paired,
+			NULL);
+
+	g_free(alias);
+}
+
+void adapter_update_found_devices(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+				int8_t rssi, uint32_t class, const char *name,
+				const char *alias, gboolean legacy,
 				name_status_t name_status)
 {
 	struct remote_dev_info *dev, match;
@@ -2544,46 +2645,33 @@ int adapter_add_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 	bacpy(&match.bdaddr, bdaddr);
 	match.name_status = NAME_ANY;
 
-	/* ignore repeated entries */
 	dev = adapter_search_found_devices(adapter, &match);
 	if (dev) {
-		/* device found, update the attributes */
-		if (rssi != 0)
-			dev->rssi = rssi;
-
-		dev->class = class;
-
-		if (alias) {
-			g_free(dev->alias);
-			dev->alias = g_strdup(alias);
-		}
-
-		/* Get remote name can be received while inquiring.  Keep in
-		 * mind that multiple inquiry result events can be received
-		 * from the same remote device.
-		 */
-		if (name_status != NAME_NOT_REQUIRED)
-			dev->name_status = name_status;
-
-		adapter->found_devices = g_slist_sort(adapter->found_devices,
-						(GCompareFunc) dev_rssi_cmp);
-
-		return -EALREADY;
+		if (rssi == dev->rssi)
+			return;
+		goto done;
 	}
 
 	dev = g_new0(struct remote_dev_info, 1);
 
 	bacpy(&dev->bdaddr, bdaddr);
-	dev->rssi = rssi;
 	dev->class = class;
+	if (name)
+		dev->name = g_strdup(name);
 	if (alias)
 		dev->alias = g_strdup(alias);
+	dev->legacy = legacy;
 	dev->name_status = name_status;
 
-	adapter->found_devices = g_slist_insert_sorted(adapter->found_devices,
-						dev, (GCompareFunc) dev_rssi_cmp);
+	adapter->found_devices = g_slist_prepend(adapter->found_devices, dev);
 
-	return 0;
+done:
+	dev->rssi = rssi;
+
+	adapter->found_devices = g_slist_sort(adapter->found_devices,
+						(GCompareFunc) dev_rssi_cmp);
+
+	adapter_emit_device_found(adapter, dev);
 }
 
 int adapter_remove_found_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
@@ -2657,7 +2745,7 @@ void adapter_remove_oor_device(struct btd_adapter *adapter, char *peer_addr)
 void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 {
 	const gchar *path = adapter_get_path(adapter);
-	gboolean powered, discoverable, pairable;
+	gboolean discoverable, pairable;
 	uint8_t real_class[3];
 	int dd;
 
@@ -2669,19 +2757,16 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 	switch (scan_mode) {
 	case SCAN_DISABLED:
 		adapter->mode = MODE_OFF;
-		powered = FALSE;
 		discoverable = FALSE;
 		pairable = FALSE;
 		break;
 	case SCAN_PAGE:
 		adapter->mode = MODE_CONNECTABLE;
-		powered = TRUE;
 		discoverable = FALSE;
 		pairable = adapter->pairable;
 		break;
 	case (SCAN_PAGE | SCAN_INQUIRY):
 		adapter->mode = MODE_DISCOVERABLE;
-		powered = TRUE;
 		discoverable = TRUE;
 		pairable = adapter->pairable;
 		if (adapter->discov_timeout != 0)
@@ -2701,11 +2786,6 @@ void adapter_mode_changed(struct btd_adapter *adapter, uint8_t scan_mode)
 		return;
 	}
 
-	if (powered == FALSE)
-		emit_property_changed(connection, path,
-					ADAPTER_INTERFACE, "Powered",
-					DBUS_TYPE_BOOLEAN, &powered);
-
 	/* If page scanning gets toggled emit the Pairable property */
 	if ((adapter->scan_mode & SCAN_PAGE) != (scan_mode & SCAN_PAGE))
 		emit_property_changed(connection, adapter->path,
@@ -2819,9 +2899,9 @@ static void agent_auth_cb(struct agent *agent, DBusError *derr,
 {
 	struct service_auth *auth = user_data;
 
-	auth->cb(derr, auth->user_data);
+	device_set_authorizing(auth->device, FALSE);
 
-	g_free(auth);
+	auth->cb(derr, auth->user_data);
 }
 
 static int btd_adapter_authorize(struct btd_adapter *adapter,
@@ -2835,6 +2915,7 @@ static int btd_adapter_authorize(struct btd_adapter *adapter,
 	char address[18];
 	gboolean trusted;
 	const gchar *dev_path;
+	int err;
 
 	ba2str(dst, address);
 	device = adapter_find_device(adapter, address);
@@ -2859,7 +2940,7 @@ static int btd_adapter_authorize(struct btd_adapter *adapter,
 	agent = device_get_agent(device);
 
 	if (!agent)
-		agent =  adapter->agent;
+		agent = adapter->agent;
 
 	if (!agent)
 		return -EPERM;
@@ -2870,10 +2951,16 @@ static int btd_adapter_authorize(struct btd_adapter *adapter,
 
 	auth->cb = cb;
 	auth->user_data = user_data;
+	auth->device = device;
 
 	dev_path = device_get_path(device);
 
-	return agent_authorize(agent, dev_path, uuid, agent_auth_cb, auth);
+	err = agent_authorize(agent, dev_path, uuid, agent_auth_cb, auth, g_free);
+
+	if (err == 0)
+		device_set_authorizing(device, TRUE);
+
+	return err;
 }
 
 int btd_request_authorization(const bdaddr_t *src, const bdaddr_t *dst,
@@ -2916,6 +3003,7 @@ int btd_cancel_authorization(const bdaddr_t *src, const bdaddr_t *dst)
 	struct btd_device *device;
 	struct agent *agent;
 	char address[18];
+	int err;
 
 	if (!adapter)
 		return -EPERM;
@@ -2933,12 +3021,17 @@ int btd_cancel_authorization(const bdaddr_t *src, const bdaddr_t *dst)
 	agent = device_get_agent(device);
 
 	if (!agent)
-		agent =  adapter->agent;
+		agent = adapter->agent;
 
 	if (!agent)
 		return -EPERM;
 
-	return agent_cancel(agent);
+	err = agent_cancel(agent);
+
+	if (err == 0)
+		device_set_authorizing(device, FALSE);
+
+	return err;
 }
 
 static gchar *adapter_any_path = NULL;
